/*
This code is copyright (C) 1998 Robert O'Callahan.
This code is free for non-commercial use. Modification in all forms is permitted.
This license continues to apply to any modified versions. This license text must be
reproduced and distributed with any modified versions.
As a matter of courtesy I (Robert O'Callahan) would like to be informed of
any potentially useful modifications.
*/

#ifndef __FWD_H
#define __FWD_H

#define FWD_REMOTE_CONNECTED 0x01
#define FWD_LOCAL_CONNECTED  0x02
#define FWD_CLOSED_REMOTE_IN 0x04
#define FWD_CLOSED_REMOTE_OUT 0x08
#define FWD_CLOSED_LOCAL_IN  0x10
#define FWD_CLOSED_LOCAL_OUT 0x20

#define FWD_FILTER_REMOVE       0
#define FWD_FILTER_RETAIN       1
#define FWD_FILTER_CLOSECHANNEL 2

#define FWD_FILTER_FROM_CLIENT 0
#define FWD_FILTER_FROM_SERVER 1

/* a length == 0 means that we're killing the channel and the filter
   should deallocate */
typedef int (* FWDFilter)(void FAR * closure, int direction,
  int FAR * length, unsigned char FAR * FAR * buf);

typedef struct {
  int status;
  int remote_num;
  SOCKET local_socket;
  int request_num;

  void FAR * filter_closure;
  FWDFilter filter;
} FWDChannel;

/* Request types */
#define FWD_LOCAL_TO_REMOTE              1
#define FWD_REMOTE_TO_LOCAL              2
#define FWD_REMOTE_X11_TO_LOCAL          3

/* If 'type' is FWD_REMOTE_X11_TO_LOCAL, then from_port must be
   -1, to_port must be 6000 + screen number, and to_host must
   be the desired X server host.

   There can be no more than one spec of a given type and from_port
   at one time.
*/
typedef struct {
  int type;
  int from_port;
  char from_port_name[32];
  int to_port;
  char to_port_name[32];
  char to_host[256];
} FWDRequestSpec;

#define FWD_DELETED                      0x01

typedef struct {
  SOCKET listening_socket;

  uint32 to_host_addr;
  char FAR * to_host_hostent_buf;
  HANDLE to_host_lookup_handle;

  int num_channels;

  int status;
  FWDRequestSpec spec;
} FWDRequest;

typedef struct {
  HWND accept_wnd;
  WNDPROC old_accept_wnd_proc;
  int num_server_listening_specs;
  /* stored in sorted order */
  FWDRequestSpec FAR * server_listening_specs;
  int num_requests;
  FWDRequest FAR * requests;
  int num_channels;
  FWDChannel FAR * channels;
  uint32 * local_host_IP_numbers;
  struct _X11AuthData FAR * X11_auth_data;
  BOOL in_interactive_mode;
} FWDState;

void FWD_init(struct _TInstVar FAR * pvar);
/* This function checks to see whether the server is (or can) listen for a
   given request. Before the SSH session's prep phase, this returns true for
   all requests. After the SSH session's prep phase, this returns true only
   for requests that the server actually was told about during the prep phase. */
BOOL FWD_can_server_listen_for(struct _TInstVar FAR * pvar, FWDRequestSpec FAR * spec);
int FWD_get_num_request_specs(struct _TInstVar FAR * pvar);
void FWD_get_request_specs(struct _TInstVar FAR * pvar, FWDRequestSpec FAR * specs, int num_specs);
void FWD_set_request_specs(struct _TInstVar FAR * pvar, FWDRequestSpec FAR * specs, int num_specs);
int FWD_compare_specs(void const FAR * void_spec1, void const FAR * void_spec2);
void FWD_prep_forwarding(struct _TInstVar FAR * pvar);
void FWD_enter_interactive_mode(struct _TInstVar FAR * pvar);
void FWD_open(struct _TInstVar FAR * pvar, uint32 remote_channel_num,
  char FAR * local_hostname, int local_port, char FAR * originator, int originator_len);
void FWD_X11_open(struct _TInstVar FAR * pvar, uint32 remote_channel_num,
  char FAR * originator, int originator_len);
void FWD_confirmed_open(struct _TInstVar FAR * pvar, uint32 local_channel_num,
  uint32 remote_channel_num);
void FWD_failed_open(struct _TInstVar FAR * pvar, uint32 local_channel_num);
void FWD_received_data(struct _TInstVar FAR * pvar, uint32 local_channel_num,
  unsigned char FAR * data, int length);
void FWD_channel_input_eof(struct _TInstVar FAR * pvar, uint32 local_channel_num);
void FWD_channel_output_eof(struct _TInstVar FAR * pvar, uint32 local_channel_num);
void FWD_end(struct _TInstVar FAR * pvar);

#endif
